@@ -47,22 +47,6 @@ typedef unsigned __int64 uint64_t;
 namespace flann
 {
 
-template<typename T>
-inline T abs(T x) { return (x<0) ? -x : x; }
-
-template<>
-inline int abs<int>(int x) { return ::abs(x); }
-
-template<>
-inline float abs<float>(float x) { return fabsf(x); }
-
-template<>
-inline double abs<double>(double x) { return fabs(x); }
-
-template<>
-inline long double abs<long double>(long double x) { return fabsl(x); }
-
-
 template<typename T>
 struct Accumulator { typedef T Type; };
 template<>
@@ -79,14 +63,6 @@ template<>
 struct Accumulator<int> { typedef float Type; };
 
 
-class TrueType
-{
-};
-
-class FalseType
-{
-};
-
 
 /**
  * Squared Euclidean distance functor.
@@ -121,7 +97,34 @@ struct L2_Simple
     }
 };
 
+template<class T>
+struct L2_3D
+{
+    typedef bool is_kdtree_distance;
 
+    typedef T ElementType;
+    typedef typename Accumulator<T>::Type ResultType;
+
+    template <typename Iterator1, typename Iterator2>
+    ResultType operator()(Iterator1 a, Iterator2 b, size_t size, ResultType /*worst_dist*/ = -1) const
+    {
+        ResultType result = ResultType();        
+        ResultType diff;
+        diff = *a++ - *b++;
+        result += diff*diff;
+        diff = *a++ - *b++;
+        result += diff*diff;
+        diff = *a++ - *b++;
+        result += diff*diff;        
+        return result;
+    }
+
+    template <typename U, typename V>
+    inline ResultType accum_dist(const U& a, const V& b, int) const
+    {
+        return (a-b)*(a-b);
+    }
+};
 
 /**
  * Squared Euclidean distance functor, optimized version
@@ -214,10 +217,10 @@ struct L1
 
         /* Process 4 items with each loop for efficiency. */
         while (a < lastgroup) {
-            diff0 = (ResultType)abs(a[0] - b[0]);
-            diff1 = (ResultType)abs(a[1] - b[1]);
-            diff2 = (ResultType)abs(a[2] - b[2]);
-            diff3 = (ResultType)abs(a[3] - b[3]);
+            diff0 = (ResultType)std::abs(a[0] - b[0]);
+            diff1 = (ResultType)std::abs(a[1] - b[1]);
+            diff2 = (ResultType)std::abs(a[2] - b[2]);
+            diff3 = (ResultType)std::abs(a[3] - b[3]);
             result += diff0 + diff1 + diff2 + diff3;
             a += 4;
             b += 4;
@@ -228,7 +231,7 @@ struct L1
         }
         /* Process last 0-3 pixels.  Not needed for standard vector lengths. */
         while (a < last) {
-            diff0 = (ResultType)abs(*a++ - *b++);
+            diff0 = (ResultType)std::abs(*a++ - *b++);
             result += diff0;
         }
         return result;
@@ -240,7 +243,7 @@ struct L1
     template <typename U, typename V>
     inline ResultType accum_dist(const U& a, const V& b, int) const
     {
-        return abs(a-b);
+        return std::abs(a-b);
     }
 };
 
@@ -277,10 +280,10 @@ struct MinkowskiDistance
 
         /* Process 4 items with each loop for efficiency. */
         while (a < lastgroup) {
-            diff0 = (ResultType)abs(a[0] - b[0]);
-            diff1 = (ResultType)abs(a[1] - b[1]);
-            diff2 = (ResultType)abs(a[2] - b[2]);
-            diff3 = (ResultType)abs(a[3] - b[3]);
+            diff0 = (ResultType)std::abs(a[0] - b[0]);
+            diff1 = (ResultType)std::abs(a[1] - b[1]);
+            diff2 = (ResultType)std::abs(a[2] - b[2]);
+            diff3 = (ResultType)std::abs(a[3] - b[3]);
             result += pow(diff0,order) + pow(diff1,order) + pow(diff2,order) + pow(diff3,order);
             a += 4;
             b += 4;
@@ -291,7 +294,7 @@ struct MinkowskiDistance
         }
         /* Process last 0-3 pixels.  Not needed for standard vector lengths. */
         while (a < last) {
-            diff0 = (ResultType)abs(*a++ - *b++);
+            diff0 = (ResultType)std::abs(*a++ - *b++);
             result += pow(diff0,order);
         }
         return result;
@@ -303,7 +306,7 @@ struct MinkowskiDistance
     template <typename U, typename V>
     inline ResultType accum_dist(const U& a, const V& b, int) const
     {
-        return pow(static_cast<ResultType>(abs(a-b)),order);
+        return pow(static_cast<ResultType>(std::abs(a-b)),order);
     }
 };
 
@@ -332,10 +335,10 @@ struct MaxDistance
 
         /* Process 4 items with each loop for efficiency. */
         while (a < lastgroup) {
-            diff0 = abs(a[0] - b[0]);
-            diff1 = abs(a[1] - b[1]);
-            diff2 = abs(a[2] - b[2]);
-            diff3 = abs(a[3] - b[3]);
+            diff0 = std::abs(a[0] - b[0]);
+            diff1 = std::abs(a[1] - b[1]);
+            diff2 = std::abs(a[2] - b[2]);
+            diff3 = std::abs(a[3] - b[3]);
             if (diff0>result) {result = diff0; }
             if (diff1>result) {result = diff1; }
             if (diff2>result) {result = diff2; }
@@ -349,7 +352,7 @@ struct MaxDistance
         }
         /* Process last 0-3 pixels.  Not needed for standard vector lengths. */
         while (a < last) {
-            diff0 = abs(*a++ - *b++);
+            diff0 = std::abs(*a++ - *b++);
             result = (diff0>result) ? diff0 : result;
         }
         return result;
@@ -545,7 +548,7 @@ struct Hamming
     }
 
     template <typename Iterator1, typename Iterator2>
-    ResultType operator()(Iterator1 a, Iterator2 b, size_t size, ResultType /*worst_dist*/ = -1) const
+    ResultType operator()(Iterator1 a, Iterator2 b, size_t size, ResultType /*worst_dist*/ = 0) const
     {
 #ifdef FLANN_PLATFORM_64_BIT
         const uint64_t* pa = reinterpret_cast<const uint64_t*>(a);
@@ -639,7 +642,7 @@ struct HellingerDistance
     typedef typename Accumulator<T>::Type ResultType;
 
     /**
-     *  Compute the histogram intersection distance
+     *  Compute the Hellinger distance
      */
     template <typename Iterator1, typename Iterator2>
     ResultType operator()(Iterator1 a, Iterator2 b, size_t size, ResultType /*worst_dist*/ = -1) const
@@ -672,7 +675,8 @@ struct HellingerDistance
     template <typename U, typename V>
     inline ResultType accum_dist(const U& a, const V& b, int) const
     {
-        return sqrt(static_cast<ResultType>(a)) - sqrt(static_cast<ResultType>(b));
+        ResultType dist = sqrt(static_cast<ResultType>(a)) - sqrt(static_cast<ResultType>(b));
+        return dist * dist;
     }
 };
 
@@ -748,7 +752,7 @@ struct KL_Divergence
         Iterator1 last = a + size;
 
         while (a < last) {
-            if (* a != 0) {
+            if ( *a != 0 && *b != 0 ) {
                 ResultType ratio = (ResultType)(*a / *b);
                 if (ratio>0) {
                     result += *a * log(ratio);
@@ -771,54 +775,16 @@ struct KL_Divergence
     inline ResultType accum_dist(const U& a, const V& b, int) const
     {
         ResultType result = ResultType();
-        ResultType ratio = (ResultType)(a / b);
-        if (ratio>0) {
-            result = a * log(ratio);
-        }
+        if( a != 0 && b != 0 ) {
+            ResultType ratio = (ResultType)(a / b);
+            if (ratio>0) {
+                result = a * log(ratio);
+            }
+		}
         return result;
     }
 };
 
-
-
-/*
- * This is a "zero iterator". It basically behaves like a zero filled
- * array to all algorithms that use arrays as iterators (STL style).
- * It's useful when there's a need to compute the distance between feature
- * and origin it and allows for better compiler optimisation than using a
- * zero-filled array.
- */
-template <typename T>
-struct ZeroIterator
-{
-
-    T operator*()
-    {
-        return 0;
-    }
-
-    T operator[](int)
-    {
-        return 0;
-    }
-
-    const ZeroIterator<T>& operator ++()
-    {
-        return *this;
-    }
-
-    ZeroIterator<T> operator ++(int)
-    {
-        return *this;
-    }
-
-    ZeroIterator<T>& operator+=(int)
-    {
-        return *this;
-    }
-
-};
-
 }
 
 #endif //FLANN_DIST_H_
