@@ -26,11 +26,15 @@
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+/* Workaround for MSVC 10, Matlab incompatibility */
+#if (_MSC_VER >= 1600)
+#include <yvals.h>
+#define __STDC_UTF_16__
+#endif
 #include <mex.h>
 #include <flann/flann.h>
 #include <stdio.h>
 #include <string.h>
-#include <sstream>
 
 
 
@@ -38,6 +42,7 @@ struct TypedIndex
 {
     flann_index_t index;
     flann_datatype_t type;
+    void* dataset;
 };
 
 
@@ -77,17 +82,42 @@ static void matlabStructToFlannStruct( const mxArray* mexParams, FLANNParameters
 
     // search
     flannParams.checks = (int)*(mxGetPr(mxGetField(mexParams, 0,"checks")));
+    flannParams.eps = (float)*(mxGetPr(mxGetField(mexParams, 0,"eps")));
+    flannParams.sorted = (int)*(mxGetPr(mxGetField(mexParams, 0,"sorted")));
+    flannParams.max_neighbors = (int)*(mxGetPr(mxGetField(mexParams, 0,"max_neighbors")));
+    flannParams.cores = (int)*(mxGetPr(mxGetField(mexParams, 0,"cores")));
+
+    // lsh
+    flannParams.table_number_ = (unsigned int)*(mxGetPr(mxGetField(mexParams, 0, "table_number")));
+    flannParams.key_size_ = (unsigned int)*(mxGetPr(mxGetField(mexParams, 0, "key_size")));
+    flannParams.multi_probe_level_ = (unsigned int)*(mxGetPr(mxGetField(mexParams, 0, "multi_probe_level")));
 }
 
-static void flannStructToMatlabStruct( const FLANNParameters& flannParams, mxArray* mexParams )
+static mxArray* flannStructToMatlabStruct( const FLANNParameters& flannParams )
 {
+    const char* fieldnames[] = {"algorithm", "checks", "eps", "sorted", "max_neighbors", "cores", "trees", "leaf_max_size", "branching", "iterations", "centers_init", "cb_index", "table_number", "key_size", "multi_probe_level"};
+    mxArray* mexParams = mxCreateStructMatrix(1, 1, sizeof(fieldnames)/sizeof(const char*), fieldnames);
+
     mxSetField(mexParams, 0, "algorithm", to_mx_array(flannParams.algorithm));
     mxSetField(mexParams, 0, "checks", to_mx_array(flannParams.checks));
+    mxSetField(mexParams, 0, "eps", to_mx_array(flannParams.eps));
+    mxSetField(mexParams, 0, "sorted", to_mx_array(flannParams.sorted));
+    mxSetField(mexParams, 0, "max_neighbors", to_mx_array(flannParams.max_neighbors));
+    mxSetField(mexParams, 0, "cores", to_mx_array(flannParams.cores));
+
     mxSetField(mexParams, 0, "trees", to_mx_array(flannParams.trees));
+    mxSetField(mexParams, 0, "leaf_max_size", to_mx_array(flannParams.trees));
+    
     mxSetField(mexParams, 0, "branching", to_mx_array(flannParams.branching));
     mxSetField(mexParams, 0, "iterations", to_mx_array(flannParams.iterations));
     mxSetField(mexParams, 0, "centers_init", to_mx_array(flannParams.centers_init));
     mxSetField(mexParams, 0, "cb_index", to_mx_array(flannParams.cb_index));
+
+    mxSetField(mexParams, 0, "table_number", to_mx_array(flannParams.table_number_));
+    mxSetField(mexParams, 0, "key_size", to_mx_array(flannParams.key_size_));
+    mxSetField(mexParams, 0, "multi_probe_level", to_mx_array(flannParams.multi_probe_level_));
+
+    return mexParams;
 }
 
 
@@ -312,6 +342,15 @@ static void _index_find_nn(int nOutArray, mxArray* OutArray[], int nInArray, con
 }
 
 
+template<typename T>
+T* copy_array(const mxArray* array)
+{
+    size_t mem_size = mxGetN(array)*mxGetM(array)*sizeof(T);
+    void* data = malloc(mem_size);
+    memcpy(data,mxGetData(array),mem_size);
+    return (T*)data;
+}
+
 /**
  * Input arguments: dataset (matrix), params (struct)
  * Output arguments: index (pointer to index), params (struct), speedup(double)
@@ -329,6 +368,8 @@ static void _build_index(int nOutArray, mxArray* OutArray[], int nInArray, const
     const mxArray* datasetMat = InArray[0];
     check_allowed_type(datasetMat);
 
+
+
     int dcount = mxGetN(datasetMat);
     int length = mxGetM(datasetMat);
 
@@ -344,24 +385,28 @@ static void _build_index(int nOutArray, mxArray* OutArray[], int nInArray, const
     TypedIndex* typedIndex = new TypedIndex();
 
     if (mxIsSingle(datasetMat)) {
-        float* dataset = (float*) mxGetData(datasetMat);
+        float* dataset = copy_array<float>(datasetMat);
         typedIndex->index = flann_build_index_float(dataset,dcount,length, &speedup, &p);
         typedIndex->type = FLANN_FLOAT32;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsDouble(datasetMat)) {
-        double* dataset = (double*) mxGetData(datasetMat);
+        double* dataset = copy_array<double>(datasetMat);
         typedIndex->index = flann_build_index_double(dataset,dcount,length, &speedup, &p);
         typedIndex->type = FLANN_FLOAT64;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsUint8(datasetMat)) {
-        unsigned char* dataset = (unsigned char*) mxGetData(datasetMat);
+        unsigned char* dataset = copy_array<unsigned char>(datasetMat);
         typedIndex->index = flann_build_index_byte(dataset,dcount,length, &speedup, &p);
         typedIndex->type = FLANN_UINT8;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsInt32(datasetMat)) {
-        int* dataset = (int*) mxGetData(datasetMat);
+        int* dataset = copy_array<int>(datasetMat);
         typedIndex->index = flann_build_index_int(dataset,dcount,length, &speedup, &p);
         typedIndex->type = FLANN_INT32;
+        typedIndex->dataset = dataset;
     }
 
     mxClassID classID;
@@ -380,9 +425,7 @@ static void _build_index(int nOutArray, mxArray* OutArray[], int nInArray, const
     pOut[0] = typedIndex;
 
     if (nOutArray > 1) {
-        const char* fieldnames[] = {"algorithm", "checks", "trees", "branching", "iterations", "centers_init", "cb_index"};
-        OutArray[1] = mxCreateStructMatrix(1, 1, sizeof(fieldnames)/sizeof(const char*), fieldnames);
-        flannStructToMatlabStruct(p, OutArray[1]);
+        OutArray[1] = flannStructToMatlabStruct(p);
     }
     if (nOutArray > 2) {
         OutArray[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
@@ -414,6 +457,7 @@ static void _free_index(int nOutArray, mxArray* OutArray[], int nInArray, const
     else if (typedIndex->type==FLANN_INT32) {
         flann_free_index_int(typedIndex->index, NULL);
     }
+    free(typedIndex->dataset);
     delete typedIndex;
 }
 
@@ -525,24 +569,28 @@ static void _load_index(int nOutArray, mxArray* OutArray[], int nInArray, const
     TypedIndex* typedIndex = new TypedIndex();
 
     if (mxIsSingle(datasetMat)) {
-        float* dataset = (float*) mxGetData(datasetMat);
+        float* dataset = copy_array<float>(datasetMat);
         typedIndex->index = flann_load_index_float(filename, dataset,dcount,length);
         typedIndex->type = FLANN_FLOAT32;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsDouble(datasetMat)) {
-        double* dataset = (double*) mxGetData(datasetMat);
+        double* dataset = copy_array<double>(datasetMat);
         typedIndex->index = flann_load_index_double(filename, dataset,dcount,length);
         typedIndex->type = FLANN_FLOAT64;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsUint8(datasetMat)) {
-        unsigned char* dataset = (unsigned char*) mxGetData(datasetMat);
+        unsigned char* dataset = copy_array<unsigned char>(datasetMat);
         typedIndex->index = flann_load_index_byte(filename, dataset,dcount,length);
         typedIndex->type = FLANN_UINT8;
+        typedIndex->dataset = dataset;
     }
     else if (mxIsInt32(datasetMat)) {
-        int* dataset = (int*) mxGetData(datasetMat);
+        int* dataset = copy_array<int>(datasetMat);
         typedIndex->index = flann_load_index_int(filename, dataset,dcount,length);
         typedIndex->type = FLANN_INT32;
+        typedIndex->dataset = dataset;
     }
 
     mxClassID classID;
@@ -582,17 +630,21 @@ static mexFunctionEntry __functionTable[] = {
 
 static void print_selector_error()
 {
-    std::ostringstream oss;
-    oss << "Expecting first argument to be one of: ";
+    char buf[512];
+    char* msg = buf;
 
+    sprintf(msg, "%s", "Expecting first argument to be one of: ");
+    msg = buf+strlen(buf);
     for (int i=0; i<sizeof(__functionTable)/sizeof(mexFunctionEntry); ++i) {
         if (i!=0) {
-            oss << ", ";
+            sprintf(msg,", ");
+            msg = buf+strlen(buf);
         }
-        oss << __functionTable[i].name;
+        sprintf(msg, "%s", __functionTable[i].name);
+        msg = buf+strlen(buf);
     }
 
-    mexErrMsgTxt(oss.str().c_str());
+    mexErrMsgTxt(buf);
 }
 
 
